#include <math.h>
#include <GL/glut.h>
#include <pthread.h>
#include <signal.h>
#ifdef WIN32
#include <windows.h>
#endif

#include "../src/grind.h"
#include "trackball.h"
#include "gl2ps.h"

/*  externals */
extern int pargc;
extern char **pargv;
extern int mode_opt;
extern pthread_t gl_tid;
extern char *modelname;

/*  local pointers to graphics stack */
static _grcall *first_of_stack, *last_of_stack;

int winh = 500, winw = 500;

/*  2d moving of the graphics, with the left mouse-button */
static boolean moving = 0;
/*  the previous mouse coord */
static double ox, oy;
/*  the resulting translation in x and y */
static double tx = 0., ty = 0.;

/*  zoom factor */
GLfloat realzoom = 1.0;
const GLfloat initzoom = 1.0;

/*  3D view transformation matrix */
static GLfloat dim3view[16];
/*  slide transformation matrix */
static GLfloat slide_matrix[16];


void startgraphics(_grcall *f) {
  static int rc = -1;

  /*  if called first time start the openGL thread */
  if (!gl_tid) {
    /*  set first element of stack */
    first_of_stack = last_of_stack = f;
    /* start graphics thread with initGL as first function call */
    if ( (rc = pthread_create(&gl_tid, NULL, initGL, NULL)) ) {
      printf("ERROR; return code from pthread_create() is %d\n", rc);
      exit(-1);
    }
  }
  else 
    return;
}

int checkgraphics() {
  return (int)gl_tid;
}

int init() {
  /*  setup trackball for mouse-driven rotation of graphics */
  tbInit(GLUT_LEFT_BUTTON);
  /*  GL_TRUE = continue spinning when mouse is released */
  /*  GL_FALSE = stop spinning when mouse is released */
  tbAnimate(GL_TRUE);

  /*  setup matrix for sliding transformation (2D movement by mouse) */
  /*  initially this is the identity matrix */
  glMatrixMode(GL_MODELVIEW);  
  glPushMatrix();
  glLoadIdentity();
  /*  store the values in our matrix */
  glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat*)slide_matrix);
  glPopMatrix();

  /*  initializing the viewing matrix used for 3D drawing */
  glPushMatrix();
  glLoadIdentity();
  /*  move away a bit */
  glTranslatef(0., 0., -5.);
  /*  rotate around the axes to give a nice view */
  /*  this should be done in one call of course */
  glRotatef(-42, 0.,1.,0.);
  glRotatef(15, 1., 0., 0.);
  glRotatef(-13.4, 0.,0.,1.);
  /*  store the values in our matrix */
  glGetFloatv(GL_MODELVIEW_MATRIX, dim3view);
  glPopMatrix();

  return 1;
}

void *initGL(void *np) {
  /*  leave positioning of the window to the window system: */
  int iwinx = -1, iwiny = -1; 

  /*  initialize openGL */
  glutInit(&pargc,pargv);
  /*  double buffer for animation by trackball, depth buffer for depth cueing */
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowPosition(iwinx, iwiny);
  glutInitWindowSize(winw, winh);
  /*  open the window */
  glutCreateWindow(modelname);

  /*  enable depth-cueing */
  glEnable(GL_DEPTH_TEST);
  /* set depth func such that objects at same depth are all drawn */
  /* (eg, when axis are drawn twice with same color 
     the last color will show) */
  glDepthFunc(GL_LEQUAL);
  /*  set clear color (black) */
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glPointSize(1.0);

  /*  set up call-back functions */
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutIdleFunc(idle);

  /*  more, related, init stuff */
  init();

  /*  enter GL mainloop */
  glutMainLoop();

  pthread_exit(NULL);
}

int drawstack(const _grcall *first, const _grcall *last) {
  const _grcall *tmp = first;

  /*  push in a new matrix */
  /*  (this poping/pushing makes shifting the image by mouse more smooth; */
  /*  don't know why) */
  glPushMatrix();
  /*  set model-view matrix */
  setMVmatrix();
  /*  draw all elements in the stack */
  drawcall(&tmp, last_of_stack);
  while(tmp != last) {
    tmp = tmp->next;
    drawcall(&tmp, last_of_stack);
  } 

  /*  finish all graphic calls */
  glFinish();
  /*  pop the matrix used */
  glPopMatrix();
  return 1;
}

void display() {
  /*  draw the stack from our own first to last pointers */
  drawstack(first_of_stack, last_of_stack);

  /*  swap front and back buffers */
  glutSwapBuffers();
}

void resetView() {
  tx = 0., ty = 0.;
  realzoom = initzoom;
  tbInit(GLUT_LEFT_BUTTON);
}

void setProjectionMatrix () {
  double zoomFactor = realzoom*0.8;
  const double zNear = 4.0, zFar = 6.0;

  float haspect = (float)winw/(float)winh;
  float vaspect = (haspect<1. ? 1./haspect : 1.0);
  haspect = (haspect<1. ? 1.0 : haspect);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  if (mode_opt == DIM3) {
    /* zoomFactor *= 4.; */
    glFrustum(-haspect*zoomFactor*(1.+tx), haspect*zoomFactor*(1.-tx), 
	      -vaspect*zoomFactor*(1.+ty), vaspect*zoomFactor*(1.-ty), 
	      zNear, zFar);
  }
  else {
    /* zoomFactor *= 4.; */
    gluOrtho2D(-haspect*zoomFactor*(1.+tx), haspect*zoomFactor*(1.-tx), 
	       -vaspect*zoomFactor*(1.+ty), vaspect*zoomFactor*(1.-ty));
  }
  glMatrixMode(GL_MODELVIEW);
}

int setMVmatrix() {
  glMatrixMode(GL_MODELVIEW);
  /* set up modelview matrix; start with identity */
  glLoadIdentity();

  /*  translation due to linear mouse movement */
  /* slideMatrix(); */

  /*  move away a bit */
  if (mode_opt == DIM3)
    glMultMatrixf((GLfloat*)dim3view);

  if (mode_opt == DIM3)
    /*  trackball rotations */
    tbMatrix();

  /*  model transformation */
  /*  look at middle of the space defined by the 2 (dim2) or 3 (dim3) axes  */
  if (mode_opt == DIM3)
    glTranslatef(-.5, -.5, -.5);
  else
    glTranslatef(-.5, -.5, 0.);
  
  return 1;
}

void getrealcoord(int x, int y, double *rx, double *ry, double *rz) {
  GLint viewport[4];
  GLdouble mvmatrix[16], projmatrix[16];
  GLint realy;          /* OpenGL y coordinate position  */

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  if(mode_opt == DIM3)
    glMultMatrixf((GLfloat*)dim3view);
  /* slideMatrix(); */
  if (mode_opt == DIM3)
    glTranslatef(-.5, -.5, -.5);
  else
    glTranslatef(-.5, -.5, 0.);
    
  glGetIntegerv (GL_VIEWPORT, viewport);
  glGetDoublev (GL_MODELVIEW_MATRIX, mvmatrix);
  glGetDoublev (GL_PROJECTION_MATRIX, projmatrix);
  glPopMatrix();


  /*  first get the proper winz value for obj_z = 0: */
  /* gluProject(0., 0., 0., mvmatrix, projmatrix, viewport, rx, ry, &realz); */
  /* printf("winz = %e\n", realz); */

  /*  note viewport[3] is height of window in pixels  */
  realy = viewport[3] - (GLint) y + 1;
  gluUnProject ((GLdouble) x, (GLdouble) realy, 0.5,
		mvmatrix, projmatrix, viewport, rx, ry, rz);
  /* printf("x = %e, y = %e, z = %e\n", *rx, *ry, *rz); */
}
 
void mouse(int button, int state, int x, int y) {
  if (mode_opt == DIM3)
    /*  trackball: */
    tbMouse(button, state, x, y);

  if (state == GLUT_DOWN)
    writemousecoord(button, state, x, y);

  /*  initiate linear, 2d movement with left mouse button */
  /*  put current mouse coords in ox and oy */
  /*  in 'motion()' we will interpret difference of x and y */
  if(button == GLUT_RIGHT_BUTTON) {
    if(state == GLUT_DOWN) {
      /*  start moving */
      moving = true;
      /*  remember x-coord and y-coord, scaled between [-1 .. 1] */
      ox = (2.0*x-(double)winw)/(double)winw;
      oy = (2.0*y-(double)winh)/(double)winh;
    }
    else {
      /*  stop moving when button is released */
      moving = false;
    }
  }
}

void motion(int x, int y) {
  GLdouble wx, wy;  /* returned world x, y, z coords */

  
  if (mode_opt == DIM3)
    /*  trackball: */
    tbMotion(x, y);
  
  /*  if moving, compute difference between previous coord and new coord */
  if (moving) {
    /*  scale coord's to [-1 .. 1] */
    wx = (2.0*x-(double)winw)/(double)winw;
    wy = (2.0*y-(double)winh)/(double)winh;
    /*  determine difference to previous coord */
    tx += wx - ox;
    ty -= wy - oy;
    /*  remember this coord */
    ox = wx;
    oy = wy;
    /*  update projection matrix which handles the sliding transformation */
    setProjectionMatrix();
    /*  remember to draw again */
    glutPostRedisplay();
  }
}

void keyboard(unsigned char key, int x, int y){
  switch(key){
  case 'r':
    resetView();
    setProjectionMatrix();
    glutPostRedisplay();
    break;
  case 'z':
    realzoom *=0.9;
    if (realzoom < .001)
      realzoom = 0.001;
    setProjectionMatrix();
    glutPostRedisplay();
    break;
  case 'Z':
    realzoom *= 1.1;
    if (realzoom > 10.0)
      realzoom = 10.;
    setProjectionMatrix();
    glutPostRedisplay();
    break;
  }
}


void reshape(int width, int height) {
  winw = width;
  winh = height;
  /*  Set the new viewport size */
  glViewport(0, 0, width, height);

  /*  also inform 'trackball';  */
  /*  not only when mode_opt==3d otherwise trackball migth be used uninitialized */
  /* tbReshape(winw, winh); */
  tbReshape(width, height);

  setProjectionMatrix();
}

void idle() {
#ifdef __unix__
  static struct timespec delay = {0, 100000};
#endif

  boolean redraw = false;
  _grcall *tmp = NULL;

  /*  first draw new things */
  redraw = check4append(&tmp, &last_of_stack);
  if (redraw) {
    /*  draw directly to front buffer */
    glDrawBuffer(GL_FRONT);
    /*  we only need to draw the new graphics */
    drawstack(tmp, last_of_stack);
    /*  finish all graphic calls */
    glFinish();
    /*  reset active buffer */
    glDrawBuffer(GL_BACK);
  }

  /*  check whether we should make postscript */
  if (checkPSdump()) {
    glutPostRedisplay();
    redraw = true;
  }

  /*  check whether the graphics stack will be erased */
  if (checkErase(first_of_stack, &last_of_stack)) {
    /*  we need to redraw the entire stack */
    glutPostRedisplay();
    return;
  }

  if (!redraw) {
    /*  if nothing happened sleep a little bit */
#ifdef __unix__
    nanosleep(&delay, NULL);
#elif defined WIN32
    Sleep(100);
#else
  ;
#endif  
  }
}
